/*
 * DANG_BEGIN_MODULE
 *
 * Description: Timer handlers for DOSEMU.
 * 
 * Maintainers: Scott Buchholz
 *
 * This code handles periodic routine scheduling for DOSEMU.  For example,
 * if you need to do an event or a class of events at a specified time or
 * set of times, create a timer and this module will take care of calling
 * you back at the appropriate time.
 *
 * DANG_END_MODULE
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 */

#include <unistd.h>

#include "dosemu.h"
#include "emu_timers.h"

/*
 * Setting TIMER_STATS to 1 will track some statistics on the timer
 * accuracy.  It will also slow things down a bit.
 */
#define TIMER_STATS  0

#if TIMER_STATS > 0
static Boolean  timer_stats = TRUE;

static u_long   timer_stat_total = 0; /* for one-shots */
static u_long   timer_stat_usecs = 0;
#endif

/*
 * The internal timer structures
 */

struct timer_struct_t {
  struct timer_struct_t  * next;
  struct timeval           pop_time;
  u_long                   usecs;
  u_long                   count;
  emu_timer_funct          func;
  emu_timer_t              pos;
#if TIMER_STATS > 0
  u_long                   stat_total;
  u_long                   stat_usecs;
#endif
};

/*
 * List of valid timers, free timer list, timer array
 */

static struct timer_struct_t * timer_head;
static struct timer_struct_t * timer_free_head;
static struct timer_struct_t * timer_free_tail;
static struct timer_struct_t   timers[EMU_MAX_TIMERS];

/*
 * These values hold the "current time"
 */
static Boolean          timer_is_curtime_valid;
static struct timeval   timer_curtime;

/*
 * Initialize the timer module (called at startup)
 *
 * This function also sets up the timer signal handler
 */
void timer_init(void)
{
  struct itimerval itv;

  timer_reset();

  itv.it_interval.tv_sec  = 0;
  itv.it_interval.tv_usec = TIMER_USECS;
  itv.it_value.tv_sec     = 0;
  itv.it_value.tv_usec    = TIMER_USECS;
  g_printf("TIMER: Requesting SIGTIMER signals every %uusecs\n", TIMER_USECS);
  setitimer(ITIMER_REAL, &itv, NULL);
}

/*
 * Reset the timer module (called on reboot)
 */
void timer_reset(void)
{
  emu_timer_t i;

  /* reset all the timers */
  timer_head      = NULL;
  timer_free_head = &timers[0];
  timer_free_tail = &timers[EMU_MAX_TIMERS-1];
  for (i = 0; i < EMU_MAX_TIMERS-1; i++) {
    timers[i].pos  = i+1;
    timers[i].next = &timers[i+1];
  }
  timer_free_tail->pos  = EMU_MAX_TIMERS;
  timer_free_tail->next = NULL;

  timer_is_curtime_valid = FALSE;

  /* reenable the timer statistics */
  timer_stats_enable();
  timer_stats_reset();
}

/*
 * Compare two timeval's
 */
static inline int timeval_comp(struct timeval a, struct timeval b)
{
  if (a.tv_sec > b.tv_sec)
    return 1;
  if (a.tv_sec < b.tv_sec)
    return -1;
  if (a.tv_usec > b.tv_usec)
    return 1;
  if (a.tv_usec < b.tv_usec)
    return -1;
  return 0;
}

/*
 * Fetch the current time
 */
static inline void timer_get_curtime(struct timeval *curtime)
{
  if (timer_is_curtime_valid)
    *curtime = timer_curtime;
  else
    gettimeofday(curtime, NULL);
}

/*
 * Adjust a "timeval" after we add a period to it
 */
static inline void timer_adjust(struct timeval *time)
{
  time->tv_sec += time->tv_usec / 1000000;
  time->tv_usec = time->tv_usec % 1000000;
}

/*
 * Add the usecs to the timeval
 */
static inline void timer_bump(struct timer_struct_t *ptr)
{
  ptr->pop_time.tv_usec += ptr->usecs;
  timer_adjust(&ptr->pop_time);
}

/*
 * Adjust a "timeval" after we add a period to it
 */
static void timer_reschedule(struct timer_struct_t * ptr)
{
  if (!timer_head || timeval_comp(ptr->pop_time, timer_head->pop_time) < 0) {
    ptr->next  = timer_head;
    timer_head = ptr;
  }
  else {
    struct timer_struct_t * temp = timer_head;
    while (temp->next) {
      if (timeval_comp(ptr->pop_time, temp->next->pop_time) < 0) {
	ptr->next  = temp->next;
	temp->next = ptr;
	break;
      }
      temp = temp->next;
    }
    if (!temp->next) {
      ptr->next  = NULL;
      temp->next = ptr;
    }
  }
}

/*
 * Handle all the events that have been scheduled
 */
void timer_schedule(struct timeval *curtime)
{
  static Boolean timer_in_schedule = FALSE;

  if (timer_head) {
    timer_is_curtime_valid = TRUE;
    timer_curtime          = *curtime;

    if (!timer_in_schedule) {
      struct timer_struct_t * ptr = timer_head;

      timer_in_schedule = TRUE;

      while (ptr && timeval_comp(ptr->pop_time, timer_curtime) < 0) {
	timer_head = ptr->next;
#if TIMER_STATS > 0
	{
	  struct timeval real_time;
	  long usecs;

	  gettimeofday(&real_time, NULL);
	  usecs = (real_time.tv_sec - ptr->pop_time.tv_sec) * 1000000 +
	          real_time.tv_usec - ptr->pop_time.tv_usec;
	  if (ptr->count) {
	    timer_stat_total++;
	    timer_stat_usecs += usecs;
	  }
	  else {
	    ptr->stat_total++;
	    ptr->stat_usecs += usecs;
	  }
	}
#endif
	timer_bump(ptr);
#ifdef CONFIG_SCOTT
	g_printf("TIMER: scheduled an event\n");
#endif
	ptr->func();
	if (ptr->count && !--ptr->count) {
	  timer_free_tail->next = ptr;
	  timer_free_tail       = ptr;
	  timer_free_tail->next = NULL;
	}
	else {
	  timer_reschedule(ptr);
	}

	ptr = timer_head;
      }

      timer_is_curtime_valid = FALSE;
      timer_in_schedule      = FALSE;
    }
  }
}

/*
 * Create a new timer popping at pop_time (or current time + inc if pop_time
 * is NULL), with period usecs_inc; 
 */
emu_timer_t timer_create(emu_timer_funct func, struct timeval *pop_time,
			 u_long usecs_inc)
{
  emu_timer_t timer = 0;
 
  if (timer_free_head != NULL) {
    struct timer_struct_t * ptr = timer_free_head;

    timer_free_head = timer_free_head->next;
    if (timer_free_head == NULL)
      timer_free_tail = NULL;

    ptr->usecs = usecs_inc;
    ptr->count = 0;
    ptr->func  = func;

    if (pop_time) {
      ptr->pop_time = *pop_time;
    }
    else {
      timer_get_curtime(&ptr->pop_time);
      timer_bump(ptr);
    }
#if TIMER_STATS > 0
    ptr->stat_total = 0;
    ptr->stat_usecs = 0;
#endif
    timer_reschedule(ptr);
    timer      = ptr->pos;
  }
  return timer;
}

/*
 * Create a one-shot timer which will pop at current time + usecs
 */
emu_timer_t timer_create_countdown(emu_timer_funct func, u_long usecs)
{
  emu_timer_t timer = 0;
 
  if (timer_free_head != NULL) {
    struct timer_struct_t * ptr = timer_free_head;

    timer_free_head = timer_free_head->next;
    if (timer_free_head == NULL)
      timer_free_tail = NULL;

    ptr->usecs = usecs;
    ptr->count = 1;
    ptr->func  = func;
    timer_get_curtime(&ptr->pop_time);
    timer_bump(ptr);
#if TIMER_STATS > 0
    ptr->stat_total = 0;
    ptr->stat_usecs = 0;
#endif
    timer_reschedule(ptr);
    timer      = ptr->pos;
  }
  return timer;
}

/*
 * Create a one-shot timer which will pop at pop_time
 */
emu_timer_t timer_create_popat(emu_timer_funct func, struct timeval pop_time)
{
  emu_timer_t timer = 0;
 
  if (timer_free_head != NULL) {
    struct timer_struct_t * ptr = timer_free_head;

    timer_free_head = timer_free_head->next;
    if (timer_free_head == NULL)
      timer_free_tail = NULL;

    ptr->usecs = 0;
    ptr->pop_time = pop_time;
    ptr->count = 1;
    ptr->func  = func;
#if TIMER_STATS > 0
    ptr->stat_total = 0;
    ptr->stat_usecs = 0;
#endif
    timer_reschedule(ptr);
    timer      = ptr->pos;
  }
  return timer;
}

/*
 * Change the timer rate and update the pop_time
 */
Boolean timer_modify_usecs(emu_timer_t timer, u_long usecs)
{
  struct timer_struct_t *ptr;

  ptr = timer_head;
  while (ptr && ptr->pos != timer)
    ptr = ptr->next;

  if (ptr) {
    ptr->usecs = usecs;
    timer_get_curtime(&ptr->pop_time);
    timer_bump(ptr);
    return TRUE;
  }
  else
    return FALSE;
}

/*
 * Enable timer statistic collection
 */
void timer_stats_enable(void)
{
#if TIMER_STATS > 0
  timer_stats = TRUE;
#endif
}

/*
 * Disable timer statistic collection
 */
void timer_stats_disable(void)
{
#if TIMER_STATS > 0
  timer_stats = FALSE;
#endif
}

/*
 * Dump the timer statistics
 */
void timer_stats_dump(void)
{
#if TIMER_STATS > 0
  struct timer_struct_t * ptr = timer_head;

  g_printf("TIMERS: timer_stats_dump:\n");
  while (ptr) {
    if (ptr->count == 0) {
      if (ptr->stat_total)
	g_printf("  timer #%d: total = %lu, avg_miss = %lu usecs\n", ptr->pos,
		 ptr->stat_total, ptr->stat_usecs/ptr->stat_total);
      else
	g_printf("  timer #%d: never popped\n", ptr->pos);
    }
    ptr = ptr->next;
  }

  if (timer_stat_total)
    g_printf("  one shots: total = %lu, avg_miss = %lu usecs\n", 
	   timer_stat_total, timer_stat_usecs/timer_stat_total);
  else
    g_printf("  one shots: never popped\n");
  g_printf("\n");
#else
  g_printf("timer stats not compiled\n");
#endif
}

/*
 * Reset the timer statistics
 */
void timer_stats_reset(void)
{
#if TIMER_STATS > 0
  struct timer_struct_t * ptr = timer_head;
  while (ptr) {
    ptr->stat_total = 0;
    ptr->stat_usecs = 0;
    ptr = ptr->next;
  }

  timer_stat_total = 0;
  timer_stat_usecs = 0;
#endif
}


#if 0
static u_long timer_div;          /* used by timer int code */
static u_long ticks_accum;        /* For timer_tick function, 100usec ticks */

/*
 * DANG_BEGIN_FUNCTION set_timer_tick_rate
 *
 * description:  ensure the 0x40 port timer is initially set correctly
 *
 * DANG_END_FUNCTION
 */
void set_timer_tick_rate(int value)
{
  ticks_accum   = 0;
  timer_div     = (value * 10000) / PIT_TICK_RATE;
  if (timer_div == 0)
    timer_div = 1;
#if 0
  i_printf("TIMERS:  interrupt rate requested %.3g Hz, granted %.3g Hz\n",
	   CLOCK_TICK_RATE/(double)value, 10000.0/timer_div);
#endif
}

/*
 * DANG_BEGIN_FUNCTION timer_tick
 *
 * description:
 *  Every time we get a TIMER signal from Linux, this procedure is called.
 *  It checks to see if we should queue a timer interrupt based on the
 *  current values.
 *
 * DANG_END_FUNCTION
 */
void timer_tick(void)
{
  struct timeval tp;
  u_long time_curr;
  static u_long time_old = 0;       /* Preserve value for next call */

  if (!config.timers) {
    h_printf("NOT CONFIG.TIMERS\n");
    return;
  }

  /* Get system time */
  gettimeofday(&tp, NULL);

  /* compute the number of 100usecs since we started */
  time_curr  = (tp.tv_sec - pit[0].time.tv_sec) * 10000;
  time_curr += (tp.tv_usec - pit[0].time.tv_usec) / 100;
  
  /* Reset old timer value to 0 if time_curr wrapped around back to 0 */
  if (time_curr < time_old) time_old = 0;
  
  /* Compute number of 100usec ticks since the last time this function ran */
  ticks_accum += (time_curr - time_old);
  
  /* Save old value of the timer */
  time_old = time_curr;

  /* Execute timer interrupt emulator handler */
  timer_int_engine();
}

/* DANG_BEGIN_FUNCTION timer_int_engine
 *
 * This is experimental TIMER-IRQ CHAIN code!
 * This is a function to determine whether it is time to invoke a
 * new timer irq 0 event.  Normally it is 18 times a second, but
 * many video games set it to 100 times per second or more.  Since
 * the kernel cannot keep an accurate timer interrupt, the job of this
 * routine is to perform a chained timer irq 0 right after the previous
 * timer irq 0.  This routine should, ideally, be called right after
 * the end of a timer irq, if possible.
 *
 * This would speed up high frequency timer interrupts if this code
 * can be converted into an assembly macro equivalent!
 *
 * DANG_END_FUNCTION
 */
void timer_int_engine(void)
{
  static continuous = 0;

  if (ticks_accum < timer_div)
    continuous = 0;                     /* Reset overflow guard */
  else {

    /* Overflow guard so that there is no more than 50 irqs in a chain */
    continuous++;
    
    if (continuous > 50) 
      continuous = 0;                   /* Reset overflow guard */
    else {
        pic_request(PIC_IRQ0);            /* Start a timer irq */
    }

    /* Decrement the ticks accumulator */
    ticks_accum -= timer_div;
    
    /* Prevent more than 250 ms of extra timer ticks from building up.
     * This ensures that if timer_int_engine get called less frequently
     * than every 250ms, the code won't choke on an overload of timer 
     * interrupts.
     */
    if (ticks_accum > 250) ticks_accum -= 250;
  }
}
#endif
