/*
 * DANG_BEGIN_MODULE
 * 
 * PHASE1 initialization:
 *
 * This is where you initialize anything necessary to create a virtual
 * PC configuration.  Intialize the cpu, the various chips (pic, pit, etc.)
 * and then setup & initialize the BIOS, etc.
 * 
 * DANG_END_MODULE
 * 
 * DANG_BEGIN_CHANGELOG
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 *
 * DANG_END_CHANGELOG
 */

#include <unistd.h>
#include <stdio.h>
#include <termios.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "dosemu.h"
#include "emu_mem.h"
#include "emu_bios.h"
#include "emu_mouse.h"
#include "emu_iodev.h"

static int first_call = 1;
/* first_call is a truly awful hack to keep some of these procedures from
   being called twice.  It should be fixed sometime. */

/*
 * DANG_BEGIN_FUNCTION map_hardware_ram
 *
 * description:
 *  Initialize the hardware direct-mapped pages
 * 
 * DANG_END_FUNCTION
 */
static inline void map_hardware_ram(void)
{
  int i, j;
  unsigned int addr, size;
  if (!config.must_spare_hardware_ram)
    return;
  open_kmem ();
  for (i = 0; i < sizeof(config.hardware_pages) - 1; i++) {
    if (config.hardware_pages[i++]) {
      j = i - 1;
      while (config.hardware_pages[i])
	i++;		/* NOTE: last byte is always ZERO */
      addr = HARDWARE_RAM_START + (j << 12);
      size = (i - j) << 12;
      if (mmap((caddr_t) addr, (size_t) size, PROT_READ | PROT_WRITE, 
	       MAP_SHARED | MAP_FIXED, mem_fd, addr) < 0) {
	error("ERROR: mmap error in map_hardware_ram %s\n", strerror (errno));
	close_kmem();
	return;
      }
      g_printf("mapped hardware ram at 0x%05x .. 0x%05x\n", addr, addr+size-1);
    }
  }
  close_kmem();
}

/* 
 * DANG_BEGIN_FUNCTION low_mem_init
 *
 * description:
 *  Initializes the lower 1Meg via mmap & sets up the HMA region
 *
 * DANG_END_FUNCTION
 */
void low_mem_init(void)
{
  char *result;
  g_printf ("DOS memory area being mapped in\n");
  result = mmap (NULL, 0x100000,
                 PROT_EXEC | PROT_READ | PROT_WRITE,
                 MAP_FIXED | MAP_PRIVATE | MAP_ANON,
                 -1, 0);
  if (result != NULL) {
    perror ("anonymous mmap");
    leavedos (1);
  }

  HMA_init();
  map_hardware_ram();
}

/* 
 * DANG_BEGIN_FUNCTION device_init
 * 
 * description:
 *  Calls all initialization routines for devices (keyboard, video, serial,
 *    disks, etc.)
 *
 * DANG_END_FUNCTION
 */
void device_init(void)
{
  if (keyboard_init() != 0) {
    error("ERROR: can't open keyboard\n");
    leavedos(19);
  }
  keyboard_flags_init();
  if (!config.vga)
    config.allowvideoportaccess = 0;
 
  scr_state_init();
  video_config_init();
  mouse_init();
}

/* 
 * DANG_BEGIN_FUNCTION bios_mem_setup
 * 
 * description:
 *  Set up all memory areas as would be present on a typical i86 during
 * the boot phase.
 *
 * DANG_END_FUNCTION
 */
static inline void bios_mem_setup(void)
{
  int b;

  keybuf_clear();

#if 0
  bios_ctrl_alt_del_flag = 0x0000;
  bios_keyboard_flags2   = 16;	/* 102-key keyboard */
#endif

  *OUTB_ADD = 1;                /* Set OUTB_ADD to 1 */
  *LASTSCAN_ADD = 1;
}

/*
 * DANG_BEGIN_FUNCTION dosemu_banner
 *
 * description:
 *  Setup the call stack to draw the dosemu banner
 * 
 * DANG_END_FUNCTION
 */
static void
dosemu_banner(void)
{
  unsigned char *ssp;
  unsigned long sp;

  ssp = (unsigned char *) (REG(ss) << 4);
  sp = (unsigned long) LWORD(esp);

g_printf("BIOS: dosemu_banner, old CS:IP = 0x%04x:0x%04x\n", _CS, _IP);

  pushw(ssp, sp, _CS);
  pushw(ssp, sp, _IP);
  _SP -= 4;
  _CS = BIOSSEG;
  _IP = ROM_BIOS_BOOTUP;
}

/*
 * DANG_BEGIN_FUNCTION phase1_init
 * 
 * description:
 *  Fire up phase 1
 *
 * DANG_END_FUNCTION
 */
void phase1_init(void)
{
  /* Virtual hardware init */
  cpu_init();		        /* initialize the CPU */
  iodev_init();                 /* initialize the io-devices */
  device_init();		/* initialize keyboard, disk, video, etc. */

  /* Memory init */
  low_mem_init();		/* initialize the lower 1Meg */

  /* BIOS setup */
  bios_init();                 /* initialize the bios */

  env_init_phase1();
  env_reset_phase1();

  dosemu_banner();

  if (config.vga) {
    g_printf("INITIALIZING VGA CARD BIOS!\n");
    init_vga_card();
  }

  /* Shared memory init */
  if (first_call) {
    shared_memory_init();
    shared_keyboard_init();
  }
  first_call = 0;

  /* Exit early check */
  if (config.exitearly) {
    dbug_printf("Leaving DOS before booting\n");
    leavedos(0);
  }
}
