/*
 * Extensions by Robert Sanders, 1992-93
 * 
 * DANG_BEGIN_MODULE
 * 
 * Here is where DOSEMU gets booted. From emu.c external calls are made to
 * the specific I/O systems (video/keyboard/serial/etc...) to initialize
 * them. Memory is cleared/set up and the boot sector is read from the
 * boot drive. Many SIGNALS are set so that DOSEMU can exploit things like
 * timers, I/O signals, illegal instructions, etc... When every system
 * gives the green light, vm86() is called to switch into vm86 mode and
 * start executing i86 code.
 * 
 * The vm86() function will return to DOSEMU when certain `exceptions` occur
 * as when some interrupt instructions occur (0xcd).
 * 
 * The top level function emulate() is called from dos.c by way of a dll
 * entry point.
 * 
 * DANG_END_MODULE
 * 
 * DANG_BEGIN_CHANGELOG
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 *
 * DANG_END_CHANGELOG
 */

/*
 * DANG_BEGIN_REMARK DOSEMU must not work within the 1 meg DOS limit, so
 * start of code is loaded at a higher address, at some time this could
 * conflict with other shared libs. If DOSEMU is compiled statically
 * (without shared libs), and org instruction is used to provide the jump
 * above 1 meg. DANG_END_REMARK
 */

#ifndef CONFIG_ELF
/*
 * DANG_BEGIN_FUNCTION jmp_emulate
 * 
 * description: This function allows the startup program `dos` to know how to
 * call the emulate function by way of the dll headers. Always make sure
 * that this line is the first of emu.c and link emu.o as the first object
 * file to the lib
 * 
 * DANG_END_FUNCTION
 */
__asm__("___START___: jmp _emulate\n");
#endif

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <getopt.h>
#include <syscall.h>
#include <signal.h>
#include <sys/vm86.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/hdreg.h>
#include <linux/fd.h>

#ifdef CONFIG_EMUMODULE
/* Please don't remove this, it's required for emusys.h (within emu.h) */
  #define __EMUSYS_parent
#endif

#include "dosemu.h"
#include "emu_bios.h"
#include "emu_termio.h"
#include "emu_video.h"
#include "emu_hma.h"
#include "emu_disks.h"
#include "emu_xms.h"
#include "emu_ipx.h"
#include "emu_serial.h"
#include "emu_cpu.h"
#include "emu_pic.h"
#include "emu_shared.h"
#include "emu_signal.h"
#include "emu_int.h"
#include "arch/signal.h"

extern void     restore_vt(unsigned short);
extern void     disallocate_vt(void);
extern void     keyboard_close(void);

void boot(void)
{
  char           *buffer;
  struct disk    *dp = NULL;

  switch (config.hdiskboot) {
  case 0:
    if (config.bootdisk)
      dp = &bootdisk;
    else
      dp = &disktab[0];
    break;
  case 1:
    dp = &hdisktab[0];
    break;
  case 2:
    dp = &disktab[1];
    break;
  default:
    error("ERROR: unexpected value for config.hdiskboot\n");
    leavedos(15);
  }
  
  ignore_segv++;
  
  disk_close();
  disk_open(dp);

  buffer = (char *) 0x7c00;
  
  if (dp->type == PARTITION) {/* we boot partition boot record, not MBR! */
    d_printf("Booting partition boot record from part=%s....\n", dp->dev_name);
    if (RPT_SYSCALL(read(dp->fdesc, buffer, SECTOR_SIZE)) != SECTOR_SIZE) {
      error("ERROR: reading partition boot sector using partition %s.\n", dp->dev_name);
      leavedos(16);
    }
  } else if (read_sectors(dp, buffer, 0, 0, 0, 1) != SECTOR_SIZE) {
    error("ERROR: can't boot from %s, using harddisk\n", dp->dev_name);
    dp = hdisktab;
    if (read_sectors(dp, buffer, 0, 0, 0, 1) != SECTOR_SIZE) {
      error("ERROR: can't boot from hard disk\n");
      leavedos(16);
    }
  }
  disk_close();
  ignore_segv--;
}

/* Silly Interrupt Generator Initialization/Closedown */

#ifdef CONFIG_SIG
SillyG_t       *SillyG = 0;
static SillyG_t SillyG_[16 + 1];
#endif

/*
 * DANG_BEGIN_FUNCTION SIG_int
 * 
 * description: Allow DOSEMU to be made aware when a hard interrupt occurs
 * Requires the sig/sillyint.o driver loaded (using NEW modules package),
 * or a kernel patch (implementing sig/int.c driver).
 * 
 * The IRQ numbers to monitor are taken from config.sillyint, each bit
 * corresponding to one IRQ. The higher 16 bit are defining the use of
 * SIGIO
 * 
 * DANG_END_FUNCTION
 */
void 
SIG_init()
{
#ifdef CONFIG_SIG
  /* Get in touch with my Silly Interrupt Driver */
  if (config.sillyint) {
    char            devname[20];
    char            prio_table[] =
      {9, 10, 11, 12, 14, 15, 3, 4, 5, 6, 7};
    int             i,
    irq,
    fd;
    SillyG_t       *sg = SillyG_;
    for (i = 0; i < sizeof(prio_table); i++) {
      irq = prio_table[i];
      if (config.sillyint & (1 << irq)) {
#ifdef CONFIG_EMUMODULE
	if (emusyscall(EMUSYS_REQUEST_IRQ, irq) < 0) {
	  g_printf("Not gonna touch IRQ %d you requested!\n", irq);
	} else {
	  g_printf("Gonna monitor the IRQ %d you requested\n", irq);
	  sg->fd = -1;
#else
	  sprintf(devname, "/dev/int/%d", irq);
	  if ((fd = open(devname, O_RDWR)) < 1) {
	    g_printf("Not gonna touch IRQ %d you requested!\n", irq);
	  } else {
	    /* Reset interupt incase it went off already */
	    RPT_SYSCALL(write(fd, NULL, (int) NULL));
	    g_printf("Gonna monitor the IRQ %d you requested, Return=0x%02x\n", irq, fd);
	    if (config.sillyint & (0x10000 << irq)) {
	      /* Use SIGIO, this should be faster */
	      add_to_io_select(fd, 1);
	    }
	    /*
	     * DANG_BEGIN_REMARK At this time we have to use
	     * SIGALRM in addition to SIGIO I don't (yet) know why
	     * the SIGIO signal gets lost sometimes (once per
	     * minute or longer). But if it happens, we can
	     * retrigger this way over SIGALRM. Normally SIGIO
	     * happens before SIGALARM, so nothing hurts. (Hans)
	     * DANG_END_REMARK
	     */
#if 0
else
#endif
  {
    /* use SIGALRM  */
    add_to_io_select(fd, 0);
  }
	    sg->fd = fd;
#endif				/* NOT CONFIG_EMUMODULE */
	    sg->irq = pic_irq_list[irq];
	    g_printf("SIG %x: enabling interrupt %x\n", irq, sg->irq);
	    pic_seti(sg->irq, do_irq, 0);
	    pic_unmaski(sg->irq);
	    sg++;
	  }
	}
      }
      sg->fd = 0;
      if (sg != SillyG_)
	SillyG = SillyG_;
    }
#endif
  }

static inline void 
SIG_close()
{
#ifdef CONFIG_SIG
  if (SillyG) {
    SillyG_t       *sg = SillyG;
#ifdef CONFIG_EMUMODULE
    while (sg->fd) {
      emusyscall(EMUSYS_FREE_IRQ, sg->irq);
      sg++;
    }
#else
    while (sg->fd)
      close((sg++)->fd);
#endif
    fprintf(stderr, "Closing all IRQ you opened!\n");
  }
#endif
}

void emumodule_init(void)
{
#ifdef CONFIG_EMUMODULE
  resolve_emusyscall();
  if (EMUSYS_AVAILABLE) {
    if (emusyscall(EMUSYS_GETVERSION, 0) < EMUSYSVERSION) {
      fprintf(stderr, "emumodule not loaded or wrong version\n\r");
      fflush(stdout);
      fflush(stderr);
      _exit(1);
    }
  }
#endif
}

/*
 * DANG_BEGIN_FUNCTION emulate
 * 
 * arguments: argc - Argument count. argv - Arguments.
 * 
 * description: Emulate gets called from dos.c. It initializes DOSEMU to
 * prepare it for running in vm86 mode. This involves catching signals,
 * preparing memory, calling all the initialization functions for the I/O
 * subsystems (video/serial/etc...), getting the boot sector instructions
 * and calling vm86().
 * 
 * DANG_END_FUNCTION
 */
#ifdef __ELF__
void main(int argc, char **argv)
#else
void emulate(int argc, char **argv)
#endif
{
  phase0_init(argc, argv);  /* set up the environment               */
  phase1_init();            /* set up the virtual PC 'box'          */
  phase2_init();            /* set up any DOS-specific features     */

  boot();	            /* read the boot sector & get moving */

  if (not_use_sigio)
    k_printf("Atleast 1 NON-SIGIO file handle in use.\n");
  else
    k_printf("No NON-SIGIO file handles in use.\n");
  g_printf("EMULATE\n");

#if 1 /* TODO -- kill me when done debugging */
  show_regs(__FILE__, __LINE__);
#endif

  fflush(stdout);
  
  while (!fatalerr) {
    run_vm86();
#if 0
    timer_int_engine();
#endif
    run_irqs();		/* trigger any hardware interrupts
			 * requested */
    serial_run();
#if 0
#ifdef CONFIG_IPX
    /* check for available packets on the packet driver interface */
    /* (timeout=0, so it immediately returns when none are available) */
    pic_request(16);
#endif
#endif
  }
  
  error("error exit: (%d,0x%04x) in_sigsegv: %d ignore_segv: %d\n",
	fatalerr, fatalerr, in_sigsegv, ignore_segv);
  
  sync();
  fprintf(stderr, "Not a good day to die!!!!!\n");
  leavedos(99);
}

static void phase1_reset(void)
{
  dbug_printf("DOS ctrl-alt-del requested.  Rebooting!\n");
  HMA_MAP(1);
  keyboard_flags_init();
  video_config_init();
  mouse_init();
  scr_state_init();
  clear_screen(READ_BYTE(BIOS_CURRENT_SCREEN_PAGE), 7);
  p_dos_str("Rebooting DOS.  Be careful...this is partially implemented\r\n");
  cpu_init();
  iodev_reset();
  env_reset_phase1();
}

static void phase2_reset(void)
{
  env_reset_phase2();
}

void dos_ctrl_alt_del(void)
{
  phase1_reset();
  phase2_reset();
  boot();
}

void dos_ctrlc(void)
{
  k_printf("DOS ctrl-c!\n");
  p_dos_str("^C\n\r");	/* print ctrl-c message */
  keybuf_clear();
  
  int_run(0x23);
}

static void ign_sigs(int sig)
{
  static int timerints = 0;
  static int otherints = 0;

  g_printf("ERROR: signal %d received in leavedos()\n", sig);
  show_regs(__FILE__, __LINE__);
  if (sig == SIG_TIME)
    timerints++;
  else
    otherints++;
  
#define LEAVEDOS_TIMEOUT (3 * 18)
#define LEAVEDOS_SIGOUT  5
  if ((timerints >= LEAVEDOS_TIMEOUT) || (otherints >= LEAVEDOS_SIGOUT)) {
    error("ERROR: timed/signalled out in leavedos()\n");
    exit(1);
  }
}

/* "graceful" shutdown */
void leavedos(int sig)
{
  struct sigaction sa;

  in_vm86 = 0;
  
  /* remove tmpdir */
  rmdir(tmpdir);
  
  setitimer(ITIMER_REAL, NULL, NULL);
  SETSIG(SIG_TIME, ign_sigs);
  SETSIG(SIGSEGV, ign_sigs);
  SETSIG(SIGILL, ign_sigs);
  SETSIG(SIGFPE, ign_sigs);
  SETSIG(SIGTRAP, ign_sigs);
  warn("leavedos(%d) called - shutting down\n", sig);
  
  g_printf("TERM: calling iodev_term()\n");
  iodev_term();

  g_printf("calling mouse_close\n");
  mouse_close();
  
#ifdef SIG
  g_printf("calling SIG_close\n");
#endif
  SIG_close();

  show_ints(0, 0x33);
  g_printf("calling video_close\n");
  video_close();
  g_printf("calling keyboard_close\n");
  fflush(stderr);
  fflush(stdout);
  keyboard_close();
  
  g_printf("calling shared memory exit\n");
  shared_memory_exit();
  if (config.detach) {
    restore_vt(config.detach);
    disallocate_vt();
  }
  _exit(sig);
}

/* return status of io_perm call */
int set_ioperm(int start, int size, int flag)
{
  int             tmp;

  if (!i_am_root)
    return -1;		/* don't bother */

  tmp = ioperm(start, size, flag);
  return tmp;
}

/*
 * DANG_BEGIN_FUNCTION add_to_io_select
 * 
 * arguments: fd - File handle to add to select statment. want_sigio -
 * Specifiy whether you want SIGIO (1) if it's available, or not (0).
 * 
 * description: Add file handle to one of 2 select FDS_SET's depending on
 * whether the kernel can handle SIGIO.
 * 
 * DANG_END_FUNCTION
 */
void add_to_io_select(int new_fd, u_char want_sigio)
{
  if (use_sigio && want_sigio) {
    int             flags;
    flags = fcntl(new_fd, F_GETFL);
    fcntl(new_fd, F_SETOWN, getpid());
    fcntl(new_fd, F_SETFL, flags | use_sigio);
    FD_SET(new_fd, &fds_sigio);
    g_printf("GEN: fd=%d gets SIGIO\n", new_fd);
  } else {
    FD_SET(new_fd, &fds_no_sigio);
    g_printf("GEN: fd=%d does not get SIGIO\n", new_fd);
    not_use_sigio++;
  }
}

/*
 * DANG_BEGIN_FUNCTION remove_from_io_select
 * 
 * arguments: fd - File handle to remove from select statment. used_sigio -
 * Specifiy whether you used SIGIO (1) if it's available, or not (0).
 * 
 * description: Remove a file handle from one of 2 select FDS_SET's depending
 * on whether the kernel can handle SIGIO.
 * 
 * DANG_END_FUNCTION
 */
void remove_from_io_select(int new_fd, u_char used_sigio)
{
  if (use_sigio && used_sigio) {
    int             flags;
    flags = fcntl(new_fd, F_GETFL);
    fcntl(new_fd, F_SETOWN, NULL);
    fcntl(new_fd, F_SETFL, flags & ~(use_sigio));
    FD_CLR(new_fd, &fds_sigio);
    g_printf("GEN: fd=%d removed from select SIGIO\n", new_fd);
  } else {
    FD_CLR(new_fd, &fds_no_sigio);
    g_printf("GEN: fd=%d removed from select\n", new_fd);
    not_use_sigio++;
  }
}

#if 1
int do_ioctl(int fd, int req, int param3)
{
  int tmp;

  if (in_sighandler && in_ioctl) {
    k_printf("KBD: do_ioctl(): in ioctl %d 0x%04x 0x%04x.\nqueuing: %d 0x%04x 0x%04x\n",
	     curi.fd, curi.req, curi.param3, fd, req, param3);
    queue_ioctl(fd, req, param3);
    errno = EDEADLOCK;
#ifdef CONFIG_SYNC_ALOT
    fflush(stdout);
    sync();			/* for safety */
#endif
    return -1;
  } else {
    in_ioctl = 1;
    curi.fd = fd;
    curi.req = req;
    curi.param3 = param3;
    if (iq.queued) {
      k_printf("KBD: detected queued ioctl in do_ioctl(): %d 0x%04x 0x%04x\n",
	       iq.fd, iq.req, iq.param3);
    }
    k_printf("KBD: IOCTL fd=0x%x, req=0x%x, param3=0x%x\n", fd, req, param3);
    tmp = ioctl(fd, req, param3);
    in_ioctl = 0;
    return tmp;
  }
}

int queue_ioctl(int fd, int req, int param3)
{
  if (iq.queued) {
    error("ioctl already queued: %d 0x%04x 0x%04x\n", iq.fd, iq.req,
	  iq.param3);
    return 1;
  }
  iq.fd = fd;
  iq.req = req;
  iq.param3 = param3;
  iq.queued = 1;
  
  return 0;			/* success */
}

void do_queued_ioctl(void)
{
  if (iq.queued) {
    iq.queued = 0;
    do_ioctl(iq.fd, iq.req, iq.param3);
  }
}

void activate(int con_num)
{
  if (in_ioctl) {
    k_printf("KBD: can't ioctl for activate, in a signal handler\n");
    do_ioctl(kbd_fd, VT_ACTIVATE, con_num);
  } else
    do_ioctl(kbd_fd, VT_ACTIVATE, con_num);
}
#endif
